from datetime import datetime
from datetime import timezone
from typing import Annotated
from typing import Optional

from fastapi import APIRouter
from fastapi import Depends
from fastapi import HTTPException
from fastapi import Query
from fastapi import Request
from sqlalchemy import select
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import Session
from sqlmodel import Column
from sqlmodel import Field
from sqlmodel import SQLModel
from sqlmodel import TIMESTAMP

from ghostforge.db import get_session
from ghostforge.db import User
from ghostforge.htmljson import HtmlJson
from ghostforge.users import get_current_user

gf = APIRouter()
hj = HtmlJson()


class Ghost(SQLModel, table=True):
    id: Optional[int] = Field(primary_key=True)
    first_name: Optional[str] = Field()
    last_name: Optional[str] = Field()
    middle_name: Optional[str] = Field()
    birthdate: Optional[datetime] = Field(
        sa_column=Column(
            TIMESTAMP(timezone=True),
        )
    )


class GhostCreate(Ghost, table=False):
    pass


class GhostUpdate(Ghost, table=False):
    pass


class GhostRead(Ghost, table=False):
    pass


@gf.get("/ghosts/{ghost_id}")
@hj.html_or_json("ghost.html")
async def read_ghost(
    ghost_id: int,
    current_user: Annotated[User, Depends(get_current_user())],
    session: AsyncSession = Depends(get_session),
    request: Request = None,
):
    ghost = await session.get(Ghost, ghost_id)
    data = {"crumbs": [("ghosts", "/ghosts"), (ghost.id, False)]}
    request.state.ghostforge = data | getattr(request.state, "ghostforge", {})
    return ghost


@gf.put("/ghosts/{ghost_id}")
async def update_ghost(
    ghost_id: int,
    ghost: GhostUpdate,
    current_user: Annotated[User, Depends(get_current_user())],
    session: AsyncSession = Depends(get_session),
    request: Request = None,
):
    db_ghost = await session.get(Ghost, ghost_id)
    if db_ghost is None:
        raise HTTPException(status_code=404, detail="No ghost with that ID")
    for k, v in ghost.dict().items():
        setattr(db_ghost, k, v)
    await session.commit()
    await session.refresh(db_ghost)
    return db_ghost


@gf.post("/ghosts")
async def create_ghost(
    ghost: GhostCreate,
    current_user: Annotated[User, Depends(get_current_user())],
    session: AsyncSession = Depends(get_session),
    request: Request = None,
):
    new_ghost = Ghost.from_orm(ghost)
    ghost.birthdate = ghost.birthdate.replace(tzinfo=timezone.utc)
    print(ghost.birthdate)
    session.add(new_ghost)
    await session.commit()
    await session.refresh(new_ghost)
    return new_ghost


@gf.get("/ghosts")
async def read_users(
    offset: int = 0, limit: int = Query(default=100, lte=100), session: Session = Depends(get_session)
):
    ghosts = await session.execute(select(Ghost).offset(offset).limit(limit))
    return ghosts.scalars().all()
